"""
Tests for glrestore
"""

import os
import sys
import time
import shutil
import pytest
import importlib
import logging
import subprocess
import pandas as pd

import glrestore
import glrestore.s3_utils

"""
SET UP TESTING UTILITIES
"""
class TestingClass():
    def teardown(self):
        importlib.reload(logging)
        if os.path.isdir(self.test_dir):
            shutil.rmtree(self.test_dir)
        os.mkdir(self.test_dir)
        #os.rmdir(self.s3_testing_loc)
        importlib.reload(logging)

@pytest.fixture()
def BTO():
    """
    Basic test object
    This object makes no copies of anything; just has references and does setup / cleanup

    aws s3 cp --profile sonn ~/Programs/inStrain/test/test_data/N5_271_010G1_scaffold_min1000.fa s3://sonn-current/users/mattolm/testing_house/glrestore/N5_271_010G1_scaffold_min1000.fa --storage-class GLACIER

    aws s3 cp --profile sonn ~/Programs/inStrain/test/test_data/N5_271_010G1_scaffold_min1000.fa s3://sonn-current/users/mattolm/testing_house/glrestore/rest_N5_271_010G1_scaffold_min1000.fa --storage-class GLACIER
    aws s3api restore-object --profile sonn --bucket sonn-current --key users/mattolm/testing_house/glrestore/rest_N5_271_010G1_scaffold_min1000.fa --restore-request '{"Days":300,"GlacierJobParameters":{"Tier":"Expedited"}}'
    """
    # Set up
    self = TestingClass()

    # Specify some s3 locations
    self.s3_testing_loc = "s3://sonn-current/users/mattolm/testing_house/glrestore/temp/"

    # Specify some file locations
    self.glacerized_file_loc = "s3://sonn-current/users/mattolm/testing_house/glrestore/N5_271_010G1_scaffold_min1000.fa"
    self.non_glacerized_file_loc = "s3://sonn-highavail/testing/ng_N5_271_010G1_scaffold_min1000.fa"
    self.restored_file_loc = "s3://sonn-current/users/mattolm/testing_house/glrestore/rest_N5_271_010G1_scaffold_min1000.fa"

    self.test_dir =  os.path.join(str(os.getcwd()), 'temp_testdir/')

    self.teardown()
    yield self
    self.teardown()

def run_glrestore(cmd):
    """
    Simulate as if you're calling glrestore from the command line
    """
    import glrestore.glrestore
    from unittest.mock import patch
    with patch.object(sys, 'argv', cmd.split(" ")):
        glrestore.glrestore.main()

def upload_glacierized_file():
    cmd = "aws s3 cp ~/Programs/inStrain/test/test_data/N5_271_010G1_scaffold_min1000.fa s3://sonn-current/users/mattolm/testing_house/glrestore/N5_271_010G1_scaffold_min1000.fa --storage-class GLACIER"
    subprocess.call(cmd, shell=True)

"""
UNIT TESTS
"""

def test_object_glacerized(BTO):
    """
    test the "s3_utils.glacier_status" function
    """
    upload_glacierized_file()

    # Test ability to check non-glacierized file
    assert glrestore.s3_utils.glacier_status(BTO.non_glacerized_file_loc) == 'no-glacier'

    # Test ability to check glacierized file
    assert glrestore.s3_utils.glacier_status(BTO.glacerized_file_loc) == 'glacier-no-restore'

    # Test ability to recognize *restored* glacierized file
    assert glrestore.s3_utils.glacier_status(BTO.restored_file_loc) == 'glacier-restored'

    # Test ability to recognize *restoring* glacierized file
    cmd = "aws s3api restore-object --bucket sonn-current --key users/mattolm/testing_house/glrestore/N5_271_010G1_scaffold_min1000.fa --restore-request '{\"Days\":1,\"GlacierJobParameters\":{\"Tier\":\"Expedited\"}}'"
    subprocess.call(cmd, shell=True)
    assert glrestore.s3_utils.glacier_status(BTO.glacerized_file_loc) == 'glacier-restoring'

def test_classify_glacier_objects(BTO):
    """
    test the "s3_utils.classify_glacier_objects" function
    """
    pd.set_option('display.max_colwidth', None)
    pd.set_option('display.max_columns', None)

    upload_glacierized_file()

    db = glrestore.s3_utils.classify_glacier_objects(
        [BTO.non_glacerized_file_loc, BTO.glacerized_file_loc, BTO.restored_file_loc])

    assert len(db) == 3
    assert db['storage_class'].value_counts()['GLACIER'] == 2

    print(db)

"""
INTEGRATED TESTS
"""

def test_glrestore_1(BTO):
    """
    A simple test restoring a single object to new location
    """
    upload_glacierized_file()

    # Make sure this file starts off glacierized
    assert glrestore.s3_utils.glacier_status(BTO.glacerized_file_loc) == 'glacier-no-restore'

    cmd = f"glrestore -f {BTO.glacerized_file_loc} --debug"
    run_glrestore(cmd)

    # Make sure this file is now being restored
    assert glrestore.s3_utils.glacier_status(BTO.glacerized_file_loc) == 'glacier-restoring'




import boto3
import logging
import datetime

import pandas as pd

from collections import defaultdict

def get_boto3_client(**kwargs):
    """
    The point of this is really to handle the "profile" option when setting up boto3
    """
    # This means you already have a client made
    if 'client' in kwargs:
        return kwargs.get('client')

    # This means I'll make you a new client
    if 'profile' in kwargs:
        profile_name = kwargs.get('profile')
        session = boto3.session.Session(profile_name=profile_name)
    else:
        session = boto3.session.Session()
    return session.client("s3")

def get_bucket_key(s3_loc):
    """
    From a full s3 location, return the bucket and key
    """
    if not s3_loc.startswith('s3://'):
        raise Exception(f"{s3_loc} is not a properly formatted key")

    bucket = s3_loc.split('/')[2]
    key = '/'.join(s3_loc.split('/')[3:])

    return bucket, key

def get_object_storage_class(s3_loc, extra_info=False, **kwargs):
    """
    Return the storage class and restoring status of an s3_loc

    If "extra_info", return a dictionary including creation date, last modified date, and size
    """
    client = get_boto3_client(**kwargs)
    bucket, key = get_bucket_key(s3_loc)
    re = client.head_object(Bucket=bucket, Key=key)

    # Get the storage class. If STANDARD, StorageClass won't be in this
    if ('ResponseMetadata' in re) & ('StorageClass' not in re):
        sclass = 'STANDARD'
    else:
        sclass = re['StorageClass']

    # Get the restore status. If not restored, 'Restore' won't be in this
    if 'Restore' not in re:
        rclass = False
    else:
        r = re['Restore']
        active_restore = 'ongoing-request="true"' in r
        # if not active_restore:
        #     expiry_date = r.split('expiry-date=\"')[1]

        if active_restore:
            rclass = 'restoring'
        else:
            rclass = 'restored'

    if extra_info:
        retd = {'storage_class':sclass, 'restore_status':rclass}
        re_header = re['ResponseMetadata']['HTTPHeaders']
        for name, rname in zip(['LastModified', 'size_bytes'], ['last-modified', 'content-length']):
            if rname in re_header:
                val = re_header[rname]

                if name in ['LastModified']:
                    val = datetime.datetime.strptime(val, "%a, %d %b %Y %H:%M:%S %Z")

                retd[name] = val
        return retd

    else:
        return sclass, rclass

def glacier_status(s3_loc, **kwargs):
    """
    Check if an object is in aws s3 glacier, and if so, return True. Else, return False.
    """
    sclass, rclass = get_object_storage_class(s3_loc, **kwargs)

    # Object is not in glacier at all
    if sclass not in ['GLACIER', 'DEEP_ARCHIVE']:
        return 'no-glacier'

    # Object is in glacier with no active restored
    elif rclass is False:
        if sclass == 'GLACIER':
            return 'glacier-no-restore'
        elif sclass == 'DEEP_ARCHIVE':
            return 'deep-glacier-no-restore'
        else:
            print(f"WHAT!! {sclass}")
            assert False, sclass

    # Object is restored in glacier
    elif rclass == 'restored':
        return 'glacier-restored'

    # Object is being actively restored
    else:
        assert rclass == 'restoring'
        return 'glacier-restoring'

def classify_glacier_objects(files, **kwargs):
    """
    Classify all files
    """
    table = defaultdict(list)
    for f in files:
        table['file'].append(f)

        rdic = get_object_storage_class(f, extra_info=True, **kwargs)
        for n, v in rdic.items():
            table[n].append(v)

    db = pd.DataFrame(table)
    db['size_bytes'] = db['size_bytes'].astype(float)
    return db

def restore_file(f, **kwargs):
    """
    Restore "f" using the parameters in kwargs
    """
    client = get_boto3_client(**kwargs)

    obucket, okey = get_bucket_key(f)

    response = client.restore_object(
        Bucket=obucket,
        Key=okey,
        RestoreRequest={
            'Days': kwargs.get('days'),
            'GlacierJobParameters': {
             'Tier': str(kwargs.get('speed'))}})

    if kwargs.get('debug', False):
        logging.debug(response)


#!/usr/bin/env python
"""
Module Docstring
"""

from .version import __version__
__author__ = "Matt Olm"
__version__ = __version__
__license__ = "MIT"

import os
import sys
import boto3
import time
import copy
import argparse
import logging
import awswrangler

import glrestore.s3_utils

def main():
    """ This is executed when run from the command line """
    args = parse_args()
    RestoreController(args).main()

class RestoreController(object):
    """
    Main controller of the restore command
    """
    def __init__(self, args):
        """
        Initialize and store args
        """
        self.args = args
        self.ori_args = copy.deepcopy(args)
        self.kwargs = vars(self.args)

    def main(self):
        """
        The main controller for restore
        """
        self.parse_arguments()

        logging.debug("Figuring out what to restore")
        self.get_files_to_restore()

        logging.debug("Classifying objects to restore")
        self.classify_objects()

        logging.debug("Print status")
        self.print_status()

        logging.debug("Restoring files")
        self.restore_files()


    def parse_arguments(self):
        """
        Parse the arguments and add them this object as attributes
        """
        args = self.kwargs

        # Set up the log
        self.setup_log()

        # Set up boto3
        if 'profile' in args:
            boto3.setup_default_session(profile_name=args.get('profile'))

    def get_files_to_restore(self):
        """
        Return a list of s3 files to restore
        """
        # Get the command line argument
        base_restore = self.kwargs.get('files')

        FILES_TO_RESTORE = []
        for br in base_restore:
            FILES_TO_RESTORE += awswrangler.s3.list_objects(br)

        self.files_to_restore = FILES_TO_RESTORE

    def classify_objects(self):
        """
        Return a table listing "file", "size", and "status"
        """
        # Run the calculation
        self.file_classifications = glrestore.s3_utils.classify_glacier_objects(self.files_to_restore)

    def print_status(self):
        """
        Print status and estimated costs
        """
        debug = self.kwargs.get('debug', False)

        cdb = self.file_classifications
        logging.info(f"Identified {len(cdb)} files")

        tdb = cdb[cdb['restore_status'] != False]
        logging.info(f"Of these, {len(tdb)} are being actively restored or are already restored")

        tdb = cdb[~cdb['storage_class'].isin(['GLACIER', 'DEEP_ARCHIVE'])]
        logging.info(f"Of these, {len(tdb)} are not in glacier")

        fcdb = cdb[(cdb['restore_status'] == False) & (cdb['storage_class'].isin(['GLACIER', 'DEEP_ARCHIVE']))]
        logging.info(f"Restoring the remaining {len(fcdb)} objects will cost the following:")

        self.display_restore_costs(fcdb)

        self.files_to_restore_filtered = fcdb['file'].tolist()

        if debug:
            for f in fcdb['file'].tolist():
                logging.debug(f)

    def display_restore_costs(self, fcdb):
        """
        Print how much this is going to cost

        NOTE- YOURE TREATING EVERYTHING AS IF IT'S BEING RESTORED FROM DEEP ARCHIVE; the "standard" is actully a bit cheaper when restoring from flexible
        """
        S3_COST_PER_GB_PER_MONTH = 0.022

        TIER2REQUEST2COST = {
            'Expedited':10,
            'Standard':0.10,
            'Bulk':0.025
        }
        TIER2REQUEST2SIZE_COST = {
            'Expedited': 0.03,
            'Standard': 0.02,
            'Bulk': 0.0025
        }

        # 0) Calculate the size and number of objects to restore
        num_obs = len(fcdb)
        size_obs = sum(fcdb['size_bytes']) / 1e9

        # 1) Calculate the cost for the extra storage
        storage_cost = size_obs * self.kwargs.get('days')

        # 2) Calculate the cost for the retrival costs
        t2cs = {}
        for tier in ['Expedited', 'Standard', 'Bulk']:
            t2cs[tier] = [(num_obs / 1000) * TIER2REQUEST2COST[tier], size_obs * TIER2REQUEST2SIZE_COST[tier]]

        # Display this info
        msg = "\n$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n"
        msg += f"It will cost ${storage_cost:.2f} to restore the {size_obs:.3f}GB of data for {self.kwargs.get('days')} days\n"

        msg += '\n'
        msg += f"Additionally it will cost the following to restore {num_obs} objects totalling {size_obs:.2f}GB:\n"
        for t, d in t2cs.items():
            msg += f"{t}: ${d[0]:.2f} + ${d[1]:.2f}\n"
        msg += "\n"
        msg += f"You chose to restore at {self.kwargs.get('speed')} speed. Please quit the program now (ctrl + c) if you'd like to change that! I'll wait 5 seconds"
        msg += "\n$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n"

        logging.info(msg)
        time.sleep(5)

    def restore_files(self):
        """
        Actually do the file restoring
        """
        files_to_restore_filtered = self.files_to_restore

        for f in files_to_restore_filtered:

            glrestore.s3_utils.restore_file(f, **self.kwargs)

        logging.info(f"Restore commands finished launching")

    def setup_log(self):
        args = self.kwargs

        # Set up the log
        root = logging.getLogger()

        if args.get('debug', False):
            root.setLevel(logging.DEBUG)
        else:
            root.setLevel(logging.INFO)

        handler = logging.StreamHandler(sys.stdout)
        if args.get('debug', False):
            root.setLevel(logging.DEBUG)
        else:
            handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        root.addHandler(handler)

        # Play debug message
        logging.debug("!" * 80)
        logging.debug("Command to run was: {0}\n".format(' '.join(sys.argv)))
        logging.debug("glrestore version {0} was run \n".format(__version__))
        logging.debug("!" * 80 + '\n')


def controller(args):
    """ Main entry point of the app """
    logging.info("hello world")
    logging.info(args)

def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-f', '--files',
        help="File or files to be restored. Can include wildcards. Must start with the bucket in the format (s3://)",
        nargs='*', default=[])

    parser.add_argument(
        '-d', '--days',
        help="Number of days to restore the data to S3 for. During this time we pay for both glacier and S3 storage of the data.",
        default=7, type=int)

    parser.add_argument(
        '-s', '--speed',
        help="Speed at which to restore the data; faster is more expensive. Expedited=(1-5 min), Standard=(3-5 hr), Bulk=(12 hr)",
        default='Expedited', choices=['Expedited', 'Standard', 'Bulk'],)

    parser.add_argument(
        '--profile',
        help="AWS credential profile to use. Will use default by default")

    parser.add_argument(
        '--debug',
        help='Create debugging log file',
        default=False, action= "store_true")
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s (version {version})".format(version=__version__))

    args = parser.parse_args()
    return args

if __name__ == "__main__":
    main()




#!/usr/bin/env python

from setuptools import setup, find_packages

import glrestore
from glrestore.version import __version__

setup(name='glrestore',
      version=__version__,
      description='Easily restore objects from AWS glacier from the command line using python',
      url='https://github.com/MrOlm/glrestore',
      author='Matthew Olm',
      author_email='mattolm@gmail.com',
      license='MIT',
      #packages=['inStrain'],
      packages=find_packages(),
      python_requires='>=3.4.0',
      install_requires=[
          'awscli',
          'boto3',
          'awswrangler'
      ],
      entry_points={
            'console_scripts': [
                  'glrestore=glrestore.glrestore:main',
            ],
      },
      zip_safe=True)
